import torch.nn as nn
import numpy as np
import math



# code sourced from https://github.com/facebookresearch/ijepa/blob/main/src/models/vision_transformer.py

def interpolate_pos_embedding(x, pos_embed):
    B, npatch, D = x.shape
    if len(pos_embed.shape) == 2:
        N, _ = pos_embed.shape
    elif len(pos_embed.shape) == 3:
        _, N, _ = pos_embed.shape

    if npatch == N:
        return pos_embed
    dim = x.shape[-1]
    pos_embed = nn.functional.interpolate(
        pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
        scale_factor=math.sqrt(npatch / N),
        mode='bicubic',
    )
    pos_embed = pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
    return pos_embed

def get_2d_sincos_pos_embed(embed_dim, height, width, cls_token=False, normalize=True):
    """
    height: int of the grid height
    width: int of the grid width
    return:
    pos_embed: [height*width, embed_dim] or [1+height*width, embed_dim] (w/ or w/o cls_token)
    """
    grid_h = np.arange(height, dtype=float)
    grid_w = np.arange(width, dtype=float)
    if normalize:
        grid_h = grid_h / (height - 1)
        grid_w = grid_w / (width - 1)
    grid = np.meshgrid(grid_w, grid_h)  # here w goes first
    grid = np.stack(grid, axis=0)

    grid = grid.reshape([2, 1, height, width])
    pos_embed = get_2d_sincos_pos_embed_from_grid(embed_dim, grid)
    if cls_token:
        pos_embed = np.concatenate([np.zeros([1, embed_dim]), pos_embed], axis=0)
    return pos_embed

def get_2d_sincos_pos_embed_from_grid(embed_dim, grid):
    assert embed_dim % 2 == 0

    # use half of dimensions to encode grid_h
    emb_h = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[0])  # (H*W, D/2)
    emb_w = get_1d_sincos_pos_embed_from_grid(embed_dim // 2, grid[1])  # (H*W, D/2)

    emb = np.concatenate([emb_h, emb_w], axis=1)  # (H*W, D)
    return emb


def get_1d_sincos_pos_embed(embed_dim, length, cls_token=False, normalize=True):
    """
    length: int of the grid length
    return:
    pos_embed: [length, embed_dim] or [1+length, embed_dim] (w/ or w/o cls_token)
    """
    grid = np.arange(length, dtype=float)
    if normalize:
        grid = grid / (length - 1)
    pos_embed = get_1d_sincos_pos_embed_from_grid(embed_dim, grid)
    if cls_token:
        pos_embed = np.concatenate([np.zeros([1, embed_dim]), pos_embed], axis=0)
    return pos_embed


def get_1d_sincos_pos_embed_from_grid(embed_dim, pos):
    """
    embed_dim: output dimension for each position
    pos: a list of positions to be encoded: size (M,)
    out: (M, D)
    """
    assert embed_dim % 2 == 0
    omega = np.arange(embed_dim // 2, dtype=float)
    omega /= embed_dim / 2.
    omega = 1. / 10000**omega   # (D/2,)

    pos = pos.reshape(-1)   # (M,)
    out = np.einsum('m,d->md', pos, omega)   # (M, D/2), outer product

    emb_sin = np.sin(out)  # (M, D/2)
    emb_cos = np.cos(out)  # (M, D/2)

    emb = np.concatenate([emb_sin, emb_cos], axis=1)  # (M, D)
    return emb



import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from joetorch.nn.modules.attention import SelfAttention, CrossAttention



class EncBlock(torch.nn.Module):
    def __init__(self, in_dim, out_dim, kernel_size, stride, padding, pool=False, bn=False, dropout=0.0, actv_layer=torch.nn.SiLU(), skip_connection=False):
        super().__init__()
        self.pool = pool
        self.bn = bn
        self.skip_connection = skip_connection

        modules = [torch.nn.Conv2d(in_dim, out_dim, kernel_size, stride, padding)]
        if pool:
            modules.append(torch.nn.MaxPool2d(kernel_size=2, stride=2))
        if bn:
            modules.append(torch.nn.BatchNorm2d(out_dim))
        if dropout > 0.0:
            modules.append(torch.nn.Dropout2d(dropout))
        if actv_layer is not None:
            modules.append(actv_layer)
        self.net = torch.nn.Sequential(*modules)
    
    def forward(self, x):
        y = self.net(x)
        if self.skip_connection:
            if self.pool:
                skip = torch.nn.functional.avg_pool2d(x, kernel_size=2, stride=2)
            else:
                skip = x
            y += skip
        return y



class DecBlock(torch.nn.Module):
    def __init__(self, in_dim, out_dim, kernel_size, stride=1, padding=0, bn=False, actv_layer=torch.nn.SiLU(), dropout=0.0, scale=1.0):
        super().__init__()
        assert scale >= 1.0, "scale must be greater than or equal to 1.0"


        bn = nn.BatchNorm2d(in_dim) if bn else None
        actv_layer = actv_layer if actv_layer is not None else nn.Identity()
        dropout = nn.Dropout2d(dropout) if dropout > 0.0 else None
        convt = nn.ConvTranspose2d(in_dim, out_dim, kernel_size, stride, padding)
        upsample = nn.Upsample(scale_factor=scale, mode='bilinear') if scale > 1.0 else None

        modules = []
        if bn is not None:
            modules.append(bn)
        if actv_layer is not None:
            modules.append(actv_layer)
        if dropout is not None:
            modules.append(dropout)

        modules.append(convt)

        if upsample is not None:
            modules.append(upsample)

        self.net = torch.nn.Sequential(*modules)
    
    def forward(self, x):
        return self.net(x)



class ConvResidualBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, gn_groups: int=32, actv_fn: nn.Module = nn.SiLU()):
        super().__init__()
        self.groupnorm_1 = nn.GroupNorm(gn_groups, in_channels)
        self.conv_1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)

        self.groupnorm_2 = nn.GroupNorm(gn_groups, out_channels)
        self.conv_2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1)

        self.actv_fn = actv_fn

        if in_channels == out_channels:
            self.residual_layer = nn.Identity()
        else:
            self.residual_layer = nn.Conv2d(in_channels, out_channels, kernel_size=1, padding=0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, Channel, Height, Width)

        residual = x

        x = self.groupnorm_1(x)
        x = self.actv_fn(x)
        x = self.conv_1(x)

        x = self.groupnorm_2(x)
        x = self.actv_fn(x)
        x = self.conv_2(x)

        return x + self.residual_layer(residual)



class ConvSelfAttentionBlock(nn.Module):
    def __init__(self, in_channels: int, gn_groups: int=32):
        super().__init__()
        self.groupnorm = nn.GroupNorm(gn_groups, in_channels)
        self.attention = SelfAttention(1, in_channels)
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:

        # x: (Batch_Size, Channel, Height, Width)
        residual = x

        x = self.groupnorm(x)
        
        n, c, h, w = x.shape
        # (Batch_Size, Channel, Height, Width) -> (Batch_Size, Height*Width, Channel)
        x = x.view(n, c, h*w).permute(0, 2, 1).contiguous()

        # (Batch_Size, Height*Width, Channel) -> (Batch_Size, Height*Width, Channel)
        x = self.attention(x)

        # (Batch_Size, Height*Width, Channel) -> (Batch_Size, Channel, Height, Width)
        x = x.permute(0, 2, 1).contiguous().view(n, c, h, w)

        return x + residual



class ConvCrossAttentionBlock(nn.Module):
    def __init__(self, q_channels: int, kv_channels: int, gn_groups: int=32):
        super().__init__()
        self.q_groupnorm = nn.GroupNorm(gn_groups, q_channels)
        self.kv_groupnorm = nn.GroupNorm(gn_groups, kv_channels)
        self.attention = CrossAttention(1, q_channels, kv_channels)
    
    def forward(self, queries: torch.Tensor, keys_n_values: torch.Tensor) -> torch.Tensor:

        # x: (Batch_Size, Channel, Height, Width)
        residual = queries

        queries = self.q_groupnorm(queries)
        keys_n_values = self.kv_groupnorm(keys_n_values)
        
        n1, c1, h1, w1 = queries.shape
        n2, c2, h2, w2 = keys_n_values.shape
        assert n1 == n2, "Batch size must be the same"

        # (Batch_Size, Channel, Height, Width) -> (Batch_Size, Height*Width, Channel)
        queries = queries.view(n1, c1, h1*w1).permute(0, 2, 1).contiguous()
        keys_n_values = keys_n_values.view(n2, c2, h2*w2).permute(0, 2, 1).contiguous()

        # (Batch_Size, Height*Width, Channel) -> (Batch_Size, Height*Width, Channel)
        x = self.attention(queries, keys_n_values)

        # (Batch_Size, Height*Width, Channel) -> (Batch_Size, Channel, Height, Width)
        x = x.permute(0, 2, 1).contiguous().view(n1, c1, h1, w1)

        return x + residual





import os
import torch
from joetorch.datasets.dataset import PreloadedDataset
from typing import Optional, Union, List, Dict, Tuple
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm
import wandb
from typing import Callable


def train(
        model: torch.nn.Module,
        train_dataset: PreloadedDataset,
        optimiser: torch.optim.Optimizer,
        num_epochs: int,
        batch_size: int,
        val_dataset: Optional[PreloadedDataset] = None,
        compute_dtype: str = 'float32',
        compile: bool = False,
        epoch_hyperparams: Dict[str, any] = {},
        loss_args: Dict[str, any] = {},
        save_dir: Optional[str] = None,
        save_metric: Tuple[str, str, str] = ('minimise', 'val', 'loss'),
        save_every: Optional[int] = None,
        save_last: bool = False,
        target_networks: Optional[List[Tuple[torch.nn.Module, torch.nn.Module, torch.Tensor]]] = None,
        max_grad_norm: Optional[float] = None,
        grad_norm_depth: int = 0,
        callback: Optional[Callable] = None,
        **kwargs,
):
    """
    Train a model for a given number of epochs.

    Args:
        model: The model to train.
        train_dataset: The training dataset.
        val_dataset: The validation dataset.
        optimiser: The optimiser to use.
        num_epochs: The number of epochs to train for.
        batch_size: The batch size.
        writer: The writer to use for logging.
        mixed_precision: Whether to use mixed precision.
        epoch_hyperparams: The hyperparameters to use for each epoch.
        loss_args: Extra arguments to pass to the loss function.
        save_dir: The directory to save the model.
        save_metric: Tuple of ('minimise'/'maximise', 'train'/'val', {metric}).
        save_every: The number of epochs between saving the model.
        save_last: Whether to save the model after the last epoch.
        target_networks: List of tuples of (target_network, network, taus).
        max_grad_norm: The maximum gradient norm.
        grad_norm_depth: The depth of the gradient norm to compute.
        wandb_init: The initialisation arguments for wandb.
    """

    assert save_metric[0] in ['minimise', 'maximise']
    assert save_metric[1] in ['train', 'val']

    if compile:
        loss_fn = torch.compile(model.loss)
    else:
        loss_fn = model.loss

    compute_dtype = getattr(torch, str(compute_dtype).split('.')[-1])
    compute_device = next(model.parameters()).device
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    val_loader = None if val_dataset is None else DataLoader(val_dataset, batch_size=batch_size, shuffle=False)

    loop = tqdm(range(num_epochs), desc='Epochs', leave=False)
    best_metric = torch.inf if save_metric[0] == 'minimise' else -torch.inf
    for epoch in loop:

        # ========================== HYPERPARAMETER UPDATE ==========================

        for key, value in epoch_hyperparams.items():
            if key == 'lr':
                for param_group in optimiser.param_groups:
                    param_group['lr'] = value[epoch]
            elif key == 'wd':
                for param_group in optimiser.param_groups:
                    if param_group['weight_decay'] != 0.0:
                        param_group['weight_decay'] = value[epoch]
            
        epoch_loss_args = {}
        for key, value in loss_args.items():
            if (isinstance(value, torch.Tensor) or isinstance(value, np.ndarray) or isinstance(value, list)) and len(value) == num_epochs:
                epoch_loss_args[key] = value[epoch]
            else:
                epoch_loss_args[key] = value
                    
        # ============================ TRAINING ============================
        if train_dataset.transform is not None:
            train_dataset.update_transformed_images()

        model.train()
        epoch_train_metrics = {}
        for batch in train_loader:
            optimiser.zero_grad()
            with torch.autocast(device_type=compute_device.type, dtype=compute_dtype, enabled=compute_dtype is not None):
                if 'train_metrics' in locals():
                    del train_metrics
                    torch.cuda.empty_cache()
                train_metrics = loss_fn(batch, **epoch_loss_args)
                train_metrics['loss'].backward()

                if max_grad_norm is not None:
                    if grad_norm_depth == 0:
                        grad_norm = torch.nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
                        train_metrics[f'{model.__class__.__name__}_grad_norm'] = grad_norm
                    else:
                        def get_submodules(module, depth, prefix=''):
                            if depth == 0:
                                return [(prefix.rstrip('.'), module)]
                            
                            submods = []
                            for name, submod in module.named_children():
                                submod_name = f"{prefix}{name}."
                                submods.extend(get_submodules(submod, depth-1, submod_name))
                            return submods if submods else [(prefix.rstrip('.'), module)]

                        for name, submod in get_submodules(model, grad_norm_depth):
                            if any(p.grad is not None for p in submod.parameters()):
                                grad_norm = torch.nn.utils.clip_grad_norm_(
                                    submod.parameters(), 
                                    max_grad_norm
                                )
                                train_metrics[f'{name}_grad_norm'] = grad_norm

                optimiser.step()

            if callback is not None:
                callback('train_batch_end', locals())
            
            for key, value in train_metrics.items():
                if key not in epoch_train_metrics:
                    epoch_train_metrics[key] = []
                epoch_train_metrics[key].append(value.detach().cpu().item())

        for key, value in epoch_train_metrics.items():
            epoch_train_metrics[key] = np.mean(value)

        if wandb.run is not None:
            for key, value in epoch_train_metrics.items():
                wandb.log({f'train/{key}': value}, step=epoch+1)
            
        postfix = {key: round(value.item(), 3) for key, value in epoch_train_metrics.items()}
        # ============================ VALIDATION ============================

        if val_loader is not None:
            if val_dataset.transform is not None:
                val_dataset.update_transformed_images()

            model.eval()
            epoch_val_metrics = {}
            for batch in val_loader:
                with torch.no_grad():
                    with torch.autocast(device_type=compute_device.type, dtype=compute_dtype, enabled=compute_dtype is not None):
                        val_metrics = loss_fn(batch, **epoch_loss_args)

                if callback is not None:
                    callback('val_batch_end', locals())

                for key, value in val_metrics.items():
                    if key not in epoch_val_metrics:
                        epoch_val_metrics[key] = []
                    epoch_val_metrics[key].append(value.detach().cpu().item())
            
            # ============================ LOGGING ============================

            # average metrics over the epoch
            for key, value in epoch_val_metrics.items():
                epoch_val_metrics[key] = np.mean(value)

            if wandb.run is not None:
                for key, value in epoch_val_metrics.items():
                    wandb.log({f'val/{key}': value}, step=epoch+1)
            
            for val_key, val_value in epoch_val_metrics.items():
                val_value = round(val_value.item(), 3)
                if val_key in postfix:
                    train_value = postfix[val_key]
                    postfix[val_key] = (train_value, val_value)
                else:
                    postfix[val_key] = val_value


        loop.set_postfix(postfix)

        if callback is not None:
            callback('epoch_end', locals())

        # ============================ SAVING ============================

        if save_dir is not None:
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            metrics = epoch_val_metrics if save_metric[1] == 'val' else epoch_train_metrics
            if (save_metric[0] == 'minimise' and metrics[save_metric[2]] < best_metric) or (save_metric[0] == 'maximise' and metrics[save_metric[2]] > best_metric):
                best_metric = metrics[save_metric[2]]
                for f in os.listdir(save_dir):
                    if 'best' in f:
                        os.remove(os.path.join(save_dir, f))
                torch.save(model.state_dict(), os.path.join(save_dir, f'best_{save_metric[1]}_{save_metric[2]}_{epoch}.pth'))
            if save_every is not None and epoch % save_every == 0:
                torch.save(model.state_dict(), os.path.join(save_dir, f'checkpoint_{epoch}.pth'))
            elif save_last:
                for f in os.listdir(save_dir):
                    if 'recent' in f:
                        os.remove(os.path.join(save_dir, f))
                torch.save(model.state_dict(), os.path.join(save_dir, f'recent_{epoch}.pth'))


import torch.nn as nn

# Adapted from https://github.com/locuslab/convmixer/blob/main/convmixer.py
class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x):
        return self.fn(x) + x

class convmixer(nn.Module):
    def __init__(self, in_shape, out_shape, hidden_dim, depth, kernel_size=None, patch_size=None):
        super().__init__()

        assert len(in_shape) == 3, "in_shape must be a tuple (C, H, W)"
        if not isinstance(out_shape, int):
            assert (len(out_shape) == 3) or (len(out_shape) == 1), "out_shape must be a tuple (C, H, W) or a single integer"

        if patch_size is None:
            patch_size = max(round(7 * in_shape[-1] / 224), 1)
        if kernel_size is None:
            kernel_size = max(round(7 * (in_shape[-1] / patch_size) / 32), 3)

        self.net = nn.Sequential(
        nn.Conv2d(in_shape[0], hidden_dim, kernel_size=patch_size, stride=patch_size),
        nn.GELU(),
        nn.BatchNorm2d(hidden_dim),
        *[nn.Sequential(
                Residual(nn.Sequential(
                    nn.Conv2d(hidden_dim, hidden_dim, kernel_size, groups=hidden_dim, padding="same"),
                    nn.GELU(),
                    nn.BatchNorm2d(hidden_dim)
                )),
                nn.Conv2d(hidden_dim, hidden_dim, kernel_size=1),
                nn.GELU(),
                nn.BatchNorm2d(hidden_dim)
        ) for i in range(depth)],
        )
        self.out_proj = nn.Conv2d(hidden_dim, out_shape[0], 1) if hidden_dim != out_shape[0] else nn.Identity()
        if len(out_shape) == 1:
            self.head = nn.Sequential(
                nn.AdaptiveAvgPool2d((1,1)),
                nn.Flatten(),
            )
        elif len(out_shape) == 3:
            self.head = nn.Sequential(
                nn.AdaptiveAvgPool2d((out_shape[1], out_shape[2])),
            )

    def forward(self, x):
        x = self.net(x)
        x = self.out_proj(x)
        x = self.head(x)

        return x


from joetorch.nn.modules.fully_connected import MLP
from joetorch.nn.modules.convolutional import EncBlock, DecBlock, ConvResidualBlock, ConvSelfAttentionBlock, ConvCrossAttentionBlock
from joetorch.nn.modules.attention import SelfAttention, CrossAttention

__all__ = ['MLP', 'EncBlock', 'DecBlock', 'ConvResidualBlock', 'ConvSelfAttentionBlock', 'ConvCrossAttentionBlock', 'SelfAttention', 'CrossAttention']
